#
# Copyright 2013-2015 Bronto Software, Inc. and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import unicode_literals
from builtins import str

import collections
import re

from xml.sax.saxutils import escape as html_escape
from bs4 import BeautifulSoup

Cell = collections.namedtuple('Cell', ['type', 'rowspan', 'colspan', 'contents'])

class Converter(object):
    def __init__(self, parser):
        self._unknown_tags = set()
        self._clear = '\n\n..\n\n'

        # Regular expressions
        self._preprocess_anchors = re.compile(r'<a\s+name\s*=\s*["\']?(.+?)["\']?\s*>')
        self._post_process_empty_lines = re.compile(r'^\s+$', re.MULTILINE)
        self._post_process_compress_lines = re.compile(r'\n{3,}')
        self._whitespace_with_newline = re.compile(r'[\s\n]+')
        self._whitespace = re.compile(r'\s+')
        self._html_tag = re.compile(r'<.*?>')

        self._preprocess_entity = re.compile(r'&(nbsp|lt|gt|amp)([^;]|[\n])')
        self._parser = parser

    # --------------------------------------------------------------------------
    # ---- reST Utility Methods ----

    def _unicode(self, s):
        if isinstance(s, unicode):
            return s
        return unicode(s, 'utf8')

    def _separate(self, s):
        return u'\n\n' + s + u'\n\n'

    def _escape_inline(self, s):
        return '\\ ' + s + '\\ '

    def _inline(self, tag, s):
        # Seems fishy if our inline markup spans lines. We will instead just return
        # the string as is
        if '\n' in s:
            return s

        s = s.strip()

        if not s:
            return s

        return self._escape_inline(tag + s.strip() + tag)

    def _role(self, role, s, label=None):
        if label:
            return self._escape_inline(':%s:`%s <%s>`' % (role, label, s))
        return self._escape_inline(':%s:`%s`' % (role, s))

    def _directive(self, directive, body=None):
        header = '\n\n.. %s::\n\n' % (directive,)

        if body:
            return header + self._left_justify(body, 3) + '\n\n'
        return header + '\n'

    def _hyperlink(self, target, label):
        return self._escape_inline('`%s <%s>`_' % (label, target))

    def _listing(self, marker, items):
        items = [self._left_justify(item, len(marker) + 1) for item in items]
        items = [marker + item[len(marker):] for item in items]
        return self._separate('..') + self._separate('\n'.join(items))

    def _left_justify(self, s, indent=0):
        lines = [l.rstrip() for l in s.split('\n')]
        indents = [len(l) - len(l.lstrip()) for l in lines if l]

        if not indents:
            return s

        shift = indent - min(indents)

        if shift < 0:
            return '\n'.join(l[-shift:] for l in lines)

        prefix = ' ' * shift
        return '\n'.join(prefix + l for l in lines)

    def _compress_whitespace(self, s, replace=' ', newlines=True):
        if newlines:
            return self._whitespace_with_newline.sub(replace, s)
        return self._whitespace.sub(replace, s)

    # --------------------------------------------------------------------------
    # ---- DOM Tree Processing ----

    def _process_table_cells(self, table):
        """ Compile all the table cells.

        Returns a list of rows. The rows may have different lengths because of
        column spans.

        """

        rows = []

        for i, tr in enumerate(table.find_all('tr')):
            row = []

            for c in tr.contents:
                cell_type = getattr(c, 'name', None)

                if cell_type not in ('td', 'th'):
                    continue

                rowspan = int(c.attrs.get('rowspan', 1))
                colspan = int(c.attrs.get('colspan', 1))
                contents = self._process_children(c).strip()

                if cell_type == 'th' and i > 0:
                    contents = self._inline('**', contents)

                row.append(Cell(cell_type, rowspan, colspan, contents))

            rows.append(row)

        return rows

    def _process_table(self, node):
        rows = self._process_table_cells(node)

        if not rows:
            return ''

        table_num_columns = max(sum(c.colspan for c in row) for row in rows)

        normalized = []

        for row in rows:
            row_num_columns = sum(c.colspan for c in row)

            if row_num_columns < table_num_columns:
                cell_type = row[-1].type if row else 'td'
                row.append(Cell(cell_type, 1, table_num_columns - row_num_columns, ''))

        col_widths = [0] * table_num_columns
        row_heights = [0] * len(rows)

        for i, row in enumerate(rows):
            j = 0
            for cell in row:
                current_w = sum(col_widths[j:j + cell.colspan])
                required_w = max(len(l) for l in cell.contents.split('\n'))

                if required_w > current_w:
                    additional = required_w - current_w
                    col_widths[j] += additional - (cell.colspan - 1) * (additional // cell.colspan)
                    for jj in range(j + 1, j + cell.colspan):
                        col_widths[jj] += (additional // cell.colspan)

                current_h = row_heights[i]
                required_h = len(cell.contents.split('\n'))

                if required_h > current_h:
                    row_heights[i] = required_h

                j += cell.colspan

        row_sep = '+' + '+'.join('-' * (l + 2) for l in col_widths) + '+'
        header_sep = '+' + '+'.join('=' * (l + 2) for l in col_widths) + '+'
        lines = [row_sep]

        for i, row in enumerate(rows):
            for y in range(0, row_heights[i]):
                line = []
                j = 0
                for c in row:
                    w = sum(n + 3 for n in col_widths[j:j+c.colspan]) - 2
                    h = row_heights[i]

                    line.append('| ')
                    cell_lines = c.contents.split('\n')
                    content = cell_lines[y] if y < len(cell_lines) else ''
                    line.append(content.ljust(w))

                    j += c.colspan

                line.append('|')
                lines.append(''.join(line))

            if i == 0 and all(c.type == 'th' for c in row):
                lines.append(header_sep)
            else:
                lines.append(row_sep)

        return self._separate('\n'.join(lines))

    def _process_children(self, node):
        parts = []
        is_newline = False

        for c in node.contents:
            part = self._process(c)

            if is_newline:
                part = part.lstrip()

            if part:
                parts.append(part)
                is_newline = part.endswith('\n')

        return ''.join(parts)

    def _process_text(self, node):
        return ''.join(node.strings)

    def _process(self, node):
        if isinstance(node, str):
            return self._compress_whitespace(node)

        simple_tags = {
            'b'      : lambda s: self._inline('**', s),
            'strong' : lambda s: self._inline('**', s),
            'i'      : lambda s: self._inline('*', s),
            'em'     : lambda s: self._inline('*', s),
            'tt'     : lambda s: self._inline('``', s),
            'code'   : lambda s: self._inline('``', s),
            'h1'     : lambda s: self._inline('**', s),
            'h2'     : lambda s: self._inline('**', s),
            'h3'     : lambda s: self._inline('**', s),
            'h4'     : lambda s: self._inline('**', s),
            'h5'     : lambda s: self._inline('**', s),
            'h6'     : lambda s: self._inline('**', s),
            'sub'    : lambda s: self._role('sub', s),
            'sup'    : lambda s: self._role('sup', s),
            'hr'     : lambda s: self._separate('') # Transitions not allowed
            }

        if node.name in simple_tags:
            return simple_tags[node.name](self._process_text(node))

        if node.name == 'p':
            return self._separate(self._process_children(node).strip())

        if node.name == 'pre':
            return self._directive('parsed-literal', self._process_text(node))

        if node.name == 'a':
            if 'name' in node.attrs:
                return self._separate('.. _' + node['name'] + ':')
            if 'href' in node.attrs:
                target = node['href']
                label = self._compress_whitespace(self._process_text(node).strip('\n'))

                if target.startswith('#'):
                    return self._role('ref', target[1:], label)
                if target.startswith('@'):
                    return self._role('java:ref', target[1:], label)
                return self._hyperlink(target, label)

        if node.name == 'ul':
            items = [self._process(n) for n in node.find_all('li', recursive=False)]
            return self._listing('*', items)

        if node.name == 'ol':
            items = [self._process(n) for n in node.find_all('li', recursive=False)]
            return self._listing('#.', items)

        if node.name == 'li':
            s = self._process_children(node)
            s = s.strip()

            # If it's multiline clear the end to correcly support nested lists
            if '\n' in s:
                s = s + '\n\n'

            return s

        if node.name == 'table':
            return self._process_table(node)

        self._unknown_tags.add(node.name)

        return self._process_children(node)

    # --------------------------------------------------------------------------
    # ---- HTML Preprocessing ----

    def _preprocess_inline_javadoc_replace(self, tag, f, s):
        parts = []

        start = '{@' + tag
        start_length = len(start)

        i = s.find(start)
        j = 0

        while i != -1:
            parts.append(s[j:i])

            # Find a closing bracket such that the brackets are balanced between
            # them. This is necessary since code examples containing { and } are
            # commonly wrapped in {@code ...} tags

            try:
                j = s.find('}', i + start_length) + 1
                while s.count('{', i, j) != s.count('}', i, j):
                    j = s.index('}', j) + 1
            except ValueError:
                raise ValueError('Unbalanced {} brackets in ' + tag + ' tag')

            parts.append(f(s[i + start_length:j - 1].strip()))
            i = s.find(start, j)

        parts.append(s[j:])

        return ''.join(parts)

    def _preprocess_replace_javadoc_link(self, s):
        s = self._compress_whitespace(s)

        target = None
        label = ''

        if ' ' not in s:
            target = s
        else:
            i = s.find(' ')

            while s.count('(', 0, i) != s.count(')', 0, i):
                i = s.find(' ', i + 1)

                if i == -1:
                    i = len(s)
                    break

            target = s[:i]
            label = s[i:]

        if target[0] == '#':
            target = target[1:]

        target = target.replace('#', '.').replace(' ', '').strip()

        # Strip HTML tags from the target
        target = self._html_tag.sub('', target)

        label = label.strip()

        return '<a href="@%s">%s</a>' % (target, label)

    def _preprocess_close_anchor_tags(self, s):
        # Add closing tags to all anchors so they are better handled by the parser
        return self._preprocess_anchors.sub(r'<a name="\1"></a>', s)

    def _preprocess_fix_entities(self, s):
        return self._preprocess_entity.sub(r'&\1;\2', s)

    def _preprocess(self, s_html):
        to_tag = lambda t: lambda m: '<%s>%s</%s>' % (t, html_escape(m), t)
        s_html = self._preprocess_inline_javadoc_replace('code', to_tag('code'), s_html)
        s_html = self._preprocess_inline_javadoc_replace('literal', to_tag('span'), s_html)
        s_html = self._preprocess_inline_javadoc_replace('docRoot', lambda m: '', s_html)
        s_html = self._preprocess_inline_javadoc_replace('linkplain', self._preprocess_replace_javadoc_link, s_html)
        s_html = self._preprocess_inline_javadoc_replace('link', self._preprocess_replace_javadoc_link, s_html)

        # Make sure all anchor tags are closed
        s_html = self._preprocess_close_anchor_tags(s_html)

        # Fix up some entitities without closing ;
        s_html = self._preprocess_fix_entities(s_html)

        return s_html

    # --------------------------------------------------------------------------
    # ---- Conversion entry point ----

    def convert(self, s_html):
        if not isinstance(s_html, str):
            s_html = str(s_html, 'utf8')

        s_html = self._preprocess(s_html)

        if not s_html.strip():
            return ''

        soup = BeautifulSoup(s_html, self._parser)
        top = soup.html.body

        result = self._process_children(top)

        # Post processing
        result = self._post_process_empty_lines.sub('', result)
        result = self._post_process_compress_lines.sub('\n\n', result)
        result = result.strip()

        return result



#!/usr/bin/env python

# Copyright (c) 2012-2022. The SimGrid Team.
# All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
Usage: python generate.py nb_nodes nb_bits end_date > deployment_file.xml
"""

import sys
import random

if len(sys.argv) != 4:
    print(
        "Usage: python generate.py nb_nodes nb_bits end_date > deployment_file.xml")
    sys.exit(1)

nb_nodes = int(sys.argv[1])
nb_bits = int(sys.argv[2])
end_date = int(sys.argv[3])

max_id = 2 ** nb_bits - 1
all_ids = [0]

sys.stdout.write("<?xml version='1.0'?>\n"
                 "<!DOCTYPE platform SYSTEM \"https://simgrid.org/simgrid.dtd\">\n"
                 "<platform version=\"4\">\n  <process host=\"node-0.simgrid.org\" function=\"node\">\n"
                 "     <argument value=\"0\"/>\n     <argument value=\"%d\"/>\n  </process>\n" % end_date)

for i in range(1, nb_nodes):
    ok = False
    while not ok:
        my_id = random.randint(0, max_id)
        ok = my_id not in all_ids
    known_id = all_ids[random.randint(0, len(all_ids) - 1)]
    start_date = i * 10
    line = "  <process host=\"node-%d.simgrid.org\" function=\"node\">\n    <argument value=\"%s\"/>"\
           "\n    <argument value=\"%s\"/>\n    <argument value=\"%d\"/>\n  </process>\n" % (
               i, my_id, known_id, end_date)
    sys.stdout.write(line)
    all_ids.append(my_id)

sys.stdout.write("</platform>")



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from argparse import ArgumentParser
import sys

from simgrid import Actor, Barrier, Engine, Host, this_actor


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    parser.add_argument(
        "--actors",
        type=int,
        required=True,
        help="Number of actors to start"
    )
    return parser


def worker(barrier: Barrier):
    """ Wait on the barrier and exits.
    :param barrier: Barrier to be awaited
    """
    this_actor.info(f"Waiting on the barrier")
    barrier.wait()
    this_actor.info("Bye")


def master(actor_count: int):
    """ Create barrier with `actor_count` expected actors, spawns `actor_count - 1` workers, then wait on the barrier
    and exits.
    :param actor_count: Spawn actor_count-1 workers and do a barrier with them
    """
    barrier = Barrier(actor_count)
    workers_count = actor_count - 1
    this_actor.info(f"Spawning {workers_count} workers")
    for i in range(workers_count):
        Actor.create(f"worker-{i}", Host.by_name("Jupiter"), worker, barrier)
    this_actor.info(f"Waiting on the barrier")
    barrier.wait()
    this_actor.info("Bye")


def main():
    settings = create_parser().parse_known_args()[0]
    if settings.actors < 1:
        raise ValueError("--actors must be greater than 0")
    e = Engine(sys.argv)
    e.load_platform(settings.platform)
    Actor.create("master", Host.by_name("Tremblay"), master, settings.actors)
    e.run()


if __name__ == "__main__":
    main()



# Copyright 2021-2022. The MBI project. All rights reserved.
# This program is free software; you can redistribute it and/or modify it under the terms of the license (GNU GPL).

import re
import os
import subprocess
import MBIutils as mbi

class Tool(mbi.AbstractTool):
    def identify(self):
        return "SimGrid wrapper"

    def build(self, rootdir, cached=True):
        if cached and (os.path.exists(f"{rootdir}/builds/SimGrid/bin/smpicc") or os.path.exists('/usr/bin/simgrid-mc')):
            return

        here = os.getcwd() # Save where we were
        os.chdir(rootdir)
        # Get a GIT checkout. Either create it, or refresh it
        if os.path.exists("tools/simgrid/.git"):
            subprocess.run("cd tools/simgrid && git pull &&  cd ../..", shell=True, check=True)
        else:
            subprocess.run("rm -rf tools/simgrid && git clone --depth=1 https://framagit.org/simgrid/simgrid.git tools/simgrid", shell=True, check=True)

        # Build and install it
        os.chdir("tools/simgrid")
        subprocess.run(f"cmake -DCMAKE_INSTALL_PREFIX={rootdir}/builds/SimGrid -Denable_model-checking=ON .", shell=True, check=True)
        subprocess.run("make -j$(nproc) install VERBOSE=1", shell=True, check=True)

        # Back to our previous directory
        os.chdir(here)


    def ensure_image(self, params=None, dockerparams=None):
        mbi.AbstractTool.ensure_image(self, "-x simgrid")

    def setup(self, rootdir):
        os.environ['PATH'] = os.environ['PATH'] + ":" + rootdir + "/builds/SimGrid/bin"
        os.environ['VERBOSE'] = '1'

    def run(self, execcmd, filename, binary, num_id, timeout, batchinfo):
        cachefile = f'{binary}_{num_id}'

        if not os.path.exists("cluster.xml"):
            with open('cluster.xml', 'w') as outfile:
                outfile.write("<?xml version='1.0'?>\n")
                outfile.write("<!DOCTYPE platform SYSTEM \"https://simgrid.org/simgrid.dtd\">\n")
                outfile.write('<platform version="4.1">\n')
                outfile.write(' <cluster id="acme" prefix="node-" radical="0-99" suffix="" speed="1Gf" bw="125MBps" lat="50us"/>\n')
                outfile.write('</platform>\n')

        execcmd = execcmd.replace("mpirun", "smpirun -wrapper simgrid-mc -platform ./cluster.xml -analyze --cfg=smpi/finalization-barrier:on --cfg=smpi/list-leaks:10 --cfg=model-check/max-depth:10000")
        execcmd = execcmd.replace('${EXE}', binary)
        execcmd = execcmd.replace('$zero_buffer', "--cfg=smpi/buffering:zero")
        execcmd = execcmd.replace('$infty_buffer', "--cfg=smpi/buffering:infty")

        mbi.run_cmd(
            buildcmd=f"smpicc {filename} -trace-call-location -g -Wl,-znorelro -Wl,-znoseparate-code -o {binary}",
            execcmd=execcmd,
            cachefile=cachefile,
            filename=filename,
            binary=binary,
            timeout=timeout,
            batchinfo=batchinfo)

    def teardown(self):
        subprocess.run("find -type f -a -executable -exec rm -f {} +", shell=True, check=True) # Remove generated cruft (binary files)
        subprocess.run("rm -f smpitmp-* core", shell=True, check=True)

    def parse(self, cachefile):
        if os.path.exists(f'{cachefile}.timeout') or os.path.exists(f'logs/simgrid/{cachefile}.timeout'):
            return 'timeout'
        if not (os.path.exists(f'{cachefile}.txt') or os.path.exists(f'logs/simgrid/{cachefile}.txt')):
            return 'failure'

        with open(f'{cachefile}.txt' if os.path.exists(f'{cachefile}.txt') else f'logs/simgrid/{cachefile}.txt', 'r') as infile:
            output = infile.read()

        if re.search(r'Compilation of .*? raised an error \(retcode: ', output):
            return 'UNIMPLEMENTED'

        if re.search('MBI_MSG_RACE', output):
            return 'MBI_MSG_RACE'

        if re.search('MC is currently not supported here', output):
            return 'failure'

        if re.search('Collective communication mismatch', output):
            return 'Collective mismatch'

        if re.search('DEADLOCK DETECTED', output):
            return 'deadlock'
        if re.search('returned MPI_ERR', output):
            return 'mpierr'
        if re.search('Not yet implemented', output):
            return 'UNIMPLEMENTED'
        if re.search('CRASH IN THE PROGRAM', output):
            return 'failure'
        if re.search('Probable memory leaks in your code: SMPI detected', output):
            return 'resleak'
        if re.search('DFS exploration ended.', output):
            return 'OK'

        print(f">>>>[ INCONCLUSIVE ]>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> ({cachefile})")
        print(output)
        print("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
        return 'other'



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

   Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: @{p2pfeature}@
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: @{persfeature}@
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>


int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int dest=0, src=0;
  int stag = 0, rtag = 0;
  int buff_size = 1;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Datatype type = MPI_INT;

  @{init1}@
  @{init2}@
  if (rank == 0) {
    dest = 1; src = 1;
    @{operation1}@
    @{start1}@
    @{write1}@ /* MBIERROR1 */
    @{fini1}@
    @{free1}@
  }else if (rank == 1){
    dest = 0; src = 0;
    @{operation2}@
    @{start2}@
    @{write2}@ /* MBIERROR2 */
    @{fini2}@
    @{free2}@
  }

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for s in gen.send + gen.isend + gen.psend:
    for r in gen.irecv + gen.precv + gen.recv:
        patterns = {}
        patterns = {'s': s, 'r': r}
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['p2pfeature'] = 'Yes' if s in gen.send else 'Lacking'
        patterns['ip2pfeature'] = 'Yes' if r in gen.irecv else 'Lacking'
        patterns['persfeature'] = 'Yes' if r in gen.precv else 'Lacking'
        patterns['s'] = s
        patterns['r'] = r
        patterns['init1'] = gen.init[s]("1")
        patterns['init2'] = gen.init[r]("2")
        patterns['fini1'] = gen.fini[s]("1")
        patterns['fini2'] = gen.fini[r]("2")
        patterns['start1'] = gen.start[s]("1")
        patterns['start2'] = gen.start[r]("2")
        patterns['operation1'] = gen.operation[s]("1")
        patterns['operation2'] = gen.operation[r]("2")
        patterns['write1'] = gen.write[s]("1")
        patterns['write2'] = gen.write[r]("2")
        patterns['free1'] = gen.free[s]("1")
        patterns['free2'] = gen.free[r]("2")
        shortdesc = ' Local Concurrency with a P2P'

        # Gtenerate a message race
        if s in gen.send and r in gen.irecv + gen.precv:
            replace = patterns.copy()
            replace['shortdesc'] = shortdesc
            replace['longdesc'] = f'The message buffer in {r} is modified before the call has been completed.'
            replace['outcome'] = 'ERROR: LocalConcurrency'
            replace['errormsg'] = 'Local Concurrency with a P2P. The receive buffer in @{r}@ is modified at @{filename}@:@{line:MBIERROR2}@ whereas there is no guarantee the message has been received.'
            gen.make_file(template, f'LocalConcurrency_{r}_{s}_nok.c', replace)
        if s in gen.isend + gen.psend and r in gen.recv:
            replace = patterns.copy()
            replace['shortdesc'] = shortdesc
            replace['longdesc'] = f'The message buffer in {s} is modified before the call has been completed.'
            replace['outcome'] = 'ERROR: LocalConcurrency'
            replace['errormsg'] = 'Local Concurrency with a P2P. The send buffer in @{s}@ is modified at @{filename}@:@{line:MBIERROR1}@ whereas there is no guarantee the message has been sent.'
            gen.make_file(template, f'LocalConcurrency_{r}_{s}_nok.c', replace)
        if s in gen.isend + gen.psend and r in gen.irecv + gen.precv:
            replace = patterns.copy()
            replace['shortdesc'] = shortdesc
            replace['longdesc'] = f'The message buffer in {s} and {r} are modified before the calls have completed.'
            replace['outcome'] = 'ERROR: LocalConcurrency'
            replace['errormsg'] = 'Local Concurrency with a P2P. The message buffers in @{s}@ and @{r}@ are modified at @{filename}@:@{line:MBIERROR1}@ and @{filename}@:@{line:MBIERROR2}@ whereas there is no guarantee the calls have been completed.'
            gen.make_file(template, f'LocalConcurrency_{r}_{s}_nok.c', replace)



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example shows how to block on the completion of a set of communications.

As for the other asynchronous examples, the sender initiate all the messages it wants to send and
pack the resulting simgrid.Comm objects in a list. All messages thus occur concurrently.

The sender then loops until there is no ongoing communication. Using wait_any() ensures that the sender
will notice events as soon as they occur even if it does not follow the order of the container.

Here, finalize messages will terminate earlier because their size is 0, so they travel faster than the
other messages of this application.  As expected, the trace shows that the finalize of worker 1 is
processed before 'Message 5' that is sent to worker 0.
"""

import sys
from simgrid import Actor, Comm, Engine, Host, Mailbox, this_actor


def sender(messages_count, msg_size, receivers_count):
    # List in which we store all ongoing communications
    pending_comms = []

    # Vector of the used mailboxes
    mboxes = [Mailbox.by_name("receiver-{:d}".format(i))
              for i in range(0, receivers_count)]

    # Start dispatching all messages to receivers, in a round robin fashion
    for i in range(0, messages_count):
        content = "Message {:d}".format(i)
        mbox = mboxes[i % receivers_count]

        this_actor.info("Send '{:s}' to '{:s}'".format(content, str(mbox)))

        # Create a communication representing the ongoing communication, and store it in pending_comms
        comm = mbox.put_async(content, msg_size)
        pending_comms.append(comm)

    # Start sending messages to let the workers know that they should stop
    for i in range(0, receivers_count):
        mbox = mboxes[i]
        this_actor.info("Send 'finalize' to '{:s}'".format(str(mbox)))
        comm = mbox.put_async("finalize", 0)
        pending_comms.append(comm)

    this_actor.info("Done dispatching all messages")

    # Now that all message exchanges were initiated, wait for their completion, in order of completion.
    #
    # This loop waits for first terminating message with wait_any() and remove it with del, until all comms are
    # terminated.
    # Even in this simple example, the pending comms do not terminate in the exact same order of creation.
    while pending_comms:
        changed_pos = Comm.wait_any(pending_comms)
        del pending_comms[changed_pos]
        if changed_pos != 0:
            this_actor.info(
                "Remove the {:d}th pending comm: it terminated earlier than another comm that was initiated first."
                .format(changed_pos))

    this_actor.info("Goodbye now!")


def receiver(my_id):
    mbox = Mailbox.by_name("receiver-{:d}".format(my_id))
    this_actor.info("Wait for my first message")
    while True:
        received = mbox.get()
        this_actor.info("I got a '{:s}'.".format(received))
        if received == "finalize":
            break  # If it's a finalize message, we're done.

if __name__ == '__main__':
    e = Engine(sys.argv)

    # Load the platform description
    e.load_platform(sys.argv[1])

    Actor.create("sender", Host.by_name("Tremblay"), sender, 6, 1000000, 2)
    Actor.create("receiver", Host.by_name("Fafard"), receiver, 0)
    Actor.create("receiver", Host.by_name("Jupiter"), receiver, 1)

    e.run()



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

   Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: Lacking
  P2P!persistent: Lacking
  COLL!basic: @{collfeature}@
  COLL!nonblocking: @{icollfeature}@
  COLL!persistent: Lacking
  COLL!tools: Yes
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define buff_size 128

int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int root = 0;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Op op = MPI_SUM;
  MPI_Datatype type = MPI_INT;
  MPI_Comm newcom;
  MPI_Comm_split(MPI_COMM_WORLD, 0, nprocs - rank, &newcom);

  @{change_com}@

  int dbs = sizeof(int)*nprocs; /* Size of the dynamic buffers for alltoall and friends */
  @{init}@
  @{start}@
  @{operation}@ /* MBIERROR */
  @{fini}@
  @{free}@

  if(newcom != MPI_COMM_NULL && newcom != MPI_COMM_WORLD)
    MPI_Comm_free(&newcom);

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


# Generate code with one collective
for c in gen.coll + gen.icoll + gen.ibarrier:
    patterns = {}
    patterns = {'c': c}
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['collfeature'] = 'Yes' if c in gen.coll else 'Lacking'
    patterns['icollfeature'] = 'Yes' if c in gen.icoll + gen.ibarrier else 'Lacking'
    patterns['c'] = c
    patterns['init'] = gen.init[c]("1")
    patterns['start'] = gen.start[c]("1")
    patterns['fini'] = gen.fini[c]("1")
    patterns['free'] = gen.free[c]("1")
    patterns['operation'] = gen.operation[c]("1")

    # Generate the correct code => to remove?
    replace = patterns.copy()
    replace['shortdesc'] = 'Collective @{c}@ with correct arguments'
    replace['longdesc'] = f'All ranks in newcom call {c} with correct arguments'
    replace['outcome'] = 'OK'
    replace['errormsg'] = ''
    replace['change_com'] = '/* No error injected here */'
    gen.make_file(template, f'ParamMatching_Com_{c}_ok.c', replace)

    # Generate the incorrect communicator matching
    replace = patterns.copy()
    replace['shortdesc'] = 'Collective @{c}@ with a communicator mismatch'
    replace['longdesc'] = 'Odd ranks call the collective on newcom while even ranks call the collective on MPI_COMM_WORLD'
    replace['outcome'] = 'ERROR: CommunicatorMatching'
    replace['errormsg'] = 'Communicator mistmatch in collectives. @{c}@ at @{filename}@:@{line:MBIERROR}@ has newcom or MPI_COMM_WORLD as a communicator.'
    replace['change_com'] = 'if (rank % 2)\n    newcom = MPI_COMM_WORLD; /* MBIERROR */'
    gen.make_file(template, f'ParamMatching_Com_{c}_nok.c', replace)

    # Generate the coll with newcom=MPI_COMM_NULL
    replace = patterns.copy()
    replace['shortdesc'] = f'Collective @{c}@ with newcom=MPI_COMM_NULL'
    replace['longdesc'] = f'Collective @{c}@ with newcom=MPI_COMM_NULL'
    replace['outcome'] = 'ERROR: InvalidCommunicator'
    replace['errormsg'] = 'Invalid communicator. @{c}@ at @{filename}@:@{line:MBIERROR}@ has MPI_COMM_NULL as a communicator.'
    replace['change_com'] = 'newcom = MPI_COMM_NULL; /* MBIERROR */'
    gen.make_file(template, f'InvalidParam_ComNull_{c}_nok.c', replace)



# Copyright (c) 2006-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example shows how to simulate a non-linear resource sharing for disk
operations.

It is inspired on the paper
"Adding Storage Simulation Capacities to the SimGridToolkit: Concepts, Models, and API"
Available at : https://hal.inria.fr/hal-01197128/document

It shows how to simulate concurrent operations degrading overall performance of IO
operations (specifically the effects presented in Fig. 8 of the paper).
"""


import functools
import sys
from simgrid import Actor, Engine, NetZone, Host, Disk, this_actor


def estimate_bw(disk: Disk, n_flows: int, read: bool):
    """ Calculates the bandwidth for disk doing async operations """
    size = 100000
    cur_time = Engine.clock
    activities = [disk.read_async(size) if read else disk.write_async(
        size) for _ in range(n_flows)]

    for act in activities:
        act.wait()

    elapsed_time = Engine.clock - cur_time
    estimated_bw = float(size * n_flows) / elapsed_time
    this_actor.info("Disk: %s, concurrent %s: %d, estimated bandwidth: %f" % (
        disk.name, "read" if read else "write", n_flows, estimated_bw))


def host_runner():
    # Estimating bw for each disk and considering concurrent flows
    for n in range(1, 15, 2):
        for disk in Host.current().get_disks():
            estimate_bw(disk, n, True)
            estimate_bw(disk, n, False)


def ssd_dynamic_sharing(disk: Disk, op: str, capacity: float, n: int) -> float:
    """
    Non-linear resource callback for SSD disks

    In this case, we have measurements for some resource sharing and directly use them to return the
    correct value
    :param disk: Disk on which the operation is happening (defined by the user through the std::bind)
    :param op: read or write operation (defined by the user through the std::bind)
    :param capacity: Resource current capacity in SimGrid
    :param n: Number of activities sharing this resource
    """
    # measurements for SSD disks
    speed = {
        "write": {1: 131.},
        "read": {1: 152., 2: 161., 3: 184., 4: 197., 5: 207., 6: 215., 7: 220., 8: 224., 9: 227., 10: 231., 11: 233.,
                 12: 235., 13: 237., 14: 238., 15: 239.}
    }

    # no special bandwidth for this disk sharing N flows, just returns maximal capacity
    if n in speed[op]:
        capacity = speed[op][n]

    return capacity


def sata_dynamic_sharing(disk: Disk, capacity: float, n: int) -> float:
    """
    Non-linear resource callback for SATA disks

    In this case, the degradation for read operations is linear and we have a formula that represents it.

    :param disk: Disk on which the operation is happening (defined by the user through the std::bind)
    :param capacity: Resource current capacity in SimGrid
    :param n: Number of activities sharing this resource
    :return: New disk capacity
    """
    return 68.3 - 1.7 * n


def create_ssd_disk(host: Host, disk_name: str):
    """ Creates an SSD disk, setting the appropriate callback for non-linear resource sharing """
    disk = host.create_disk(disk_name, "240MBps", "170MBps")
    disk.set_sharing_policy(Disk.Operation.READ, Disk.SharingPolicy.NONLINEAR,
                            functools.partial(ssd_dynamic_sharing, disk, "read"))
    disk.set_sharing_policy(Disk.Operation.WRITE, Disk.SharingPolicy.NONLINEAR,
                            functools.partial(ssd_dynamic_sharing, disk, "write"))
    disk.set_sharing_policy(Disk.Operation.READWRITE,
                            Disk.SharingPolicy.LINEAR)


def create_sata_disk(host: Host, disk_name: str):
    """ Same for a SATA disk, only read operation follows a non-linear resource sharing """
    disk = host.create_disk(disk_name, "68MBps", "50MBps")
    disk.set_sharing_policy(Disk.Operation.READ, Disk.SharingPolicy.NONLINEAR,
                            functools.partial(sata_dynamic_sharing, disk))
    # this is the default behavior, expliciting only to make it clearer
    disk.set_sharing_policy(Disk.Operation.WRITE, Disk.SharingPolicy.LINEAR)
    disk.set_sharing_policy(Disk.Operation.READWRITE,
                            Disk.SharingPolicy.LINEAR)


if __name__ == '__main__':
    e = Engine(sys.argv)
    # simple platform containing 1 host and 2 disk
    zone = NetZone.create_full_zone("bob_zone")
    bob = zone.create_host("bob", 1e6)
    create_ssd_disk(bob, "Edel (SSD)")
    create_sata_disk(bob, "Griffon (SATA II)")
    zone.seal()

    Actor.create("runner", bob, host_runner)

    e.run()
    this_actor.info("Simulated time: %g" % e.clock)

    # explicitly deleting Engine object to avoid segfault during cleanup phase.
    # During Engine destruction, the cleanup of std::function linked to non_linear callback is called.
    # If we let the cleanup by itself, it fails trying on its destruction because the python main program
    # has already freed its variables
    del e



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from typing import List, Tuple
import sys

from simgrid import Engine, Actor, Comm, Host, LinkInRoute, Mailbox, NetZone, this_actor, PyGetAsync


RECEIVER_MAILBOX_NAME = "receiver"


class Sender(object):
    def __init__(self, message_size: int, messages_count: int):
        self.message_size = message_size
        self.messages_count = messages_count

    def __call__(self) -> None:
        # List in which we store all ongoing communications
        pending_comms: List[Comm] = []

        # Make a vector of the mailboxes to use
        receiver_mailbox: Mailbox = Mailbox.by_name(RECEIVER_MAILBOX_NAME)
        for i in range(self.messages_count):
            message_content = f"Message {i}"
            this_actor.info(f"Send '{message_content}' to '{receiver_mailbox.name}'")
            # Create a communication representing the ongoing communication, and store it in pending_comms
            pending_comms.append(receiver_mailbox.put_async(message_content, self.message_size))

        this_actor.info("Done dispatching all messages")

        # Now that all message exchanges were initiated, wait for their completion in one single call
        Comm.wait_all(pending_comms)

        this_actor.info("Goodbye now!")


class Receiver(object):
    def __init__(self, messages_count: int):
        self.mailbox: Mailbox = Mailbox.by_name(RECEIVER_MAILBOX_NAME)
        self.messages_count = messages_count

    def __call__(self):
        # List in which we store all incoming msgs
        pending_comms: List[Tuple[Comm, PyGetAsync]] = []
        this_actor.info(f"Wait for {self.messages_count} messages asynchronously")
        for _ in range(self.messages_count):
            pending_comms.append(self.mailbox.get_async())
        while pending_comms:
            index = Comm.wait_any([comm for (comm, _) in pending_comms])
            _, async_data = pending_comms[index]
            this_actor.info(f"I got '{async_data.get()}'.")
            pending_comms.pop(index)


def main():
    e = Engine(sys.argv)
    # Creates the platform
    #  ________                 __________
    # | Sender |===============| Receiver |
    # |________|    Link1      |__________|
    #
    zone: NetZone = NetZone.create_full_zone("Zone1")
    sender_host: Host = zone.create_host("sender", 1).seal()
    receiver_host: Host = zone.create_host("receiver", 1).seal()

    # create split-duplex link1 (UP/DOWN), limiting the number of concurrent flows in it for 2
    link = zone.create_split_duplex_link("link1", 10e9).set_latency(10e-6).set_concurrency_limit(2).seal()

    # create routes between nodes
    zone.add_route(
        sender_host.netpoint,
        receiver_host.netpoint,
        None,
        None,
        [LinkInRoute(link, LinkInRoute.UP)],
        True
    )
    zone.seal()

    # create actors Sender/Receiver
    messages_count = 10
    Actor.create("receiver", receiver_host, Receiver(messages_count=messages_count))
    Actor.create("sender", sender_host, Sender(messages_count=messages_count, message_size=int(1e6)))

    e.run()


if __name__ == "__main__":
    main()
